""""""

from enum import Enum
import pandas as pd
import numpy as np


class IATData:
    """Data model to handle the IAT data generated from the Qualtrics IAT survey"""
    mock_study = "iat"
    
    def __init__(self,
                 data_file,
                 grouped_by=("study", "ResponseId"),
                 trial_response_separator="_",
                 suffix_responses="Responses",
                 suffix_trials="Trials",
                 suffix_conditions="blockConditions",
                 congruency_labels=None):
        """Initialize the data model instance of the IATData
        :param data_file: Union[csv, zip, Bytes], the data file containing the IAT survey responses
        :param grouped_by: tuple, the indices that identify distinct IAT sessions
        :param trial_response_separator: str, the separator between responses in the embedded data
        :param suffix_responses: str, the suffix of the embedded field for saving trial responses
        :param suffix_trials: str, the suffix of the embedded field for saving trial stimuli
        :param suffix_conditions: str, the suffix of the embedded field for saving the block conditions
        :param congruency_labels: Union[None, dict], the labels by congruency
        :return None
        """
        self.suffix_responses, self.suffix_trials, self.suffix_conditions = \
            suffices = suffix_responses, suffix_trials, suffix_conditions
        self.grouped_by = grouped_by
        if not congruency_labels:
            congruency_labels = {"con": ("p+", "n-"), "inc": ("p-", "n+")}
        self.congruency_labels = congruency_labels
        self.trial_response_separator = trial_response_separator
        self.iat_data_clean = None
        
        _iat_data = pd.read_csv(data_file)
        self.iat_data = _iat_data[_iat_data[grouped_by[1]].str.startswith('R_')].reset_index(drop=True)
        
        responses_field_name = f"block1{suffix_responses}"
        self.studies = {x.split("_")[0] for x in self.iat_data.columns
                        if len(x) > len(responses_field_name) and x.endswith(responses_field_name)}
        # Create a mock study to streamline the data processing process as if there are multiple studies
        if not self.studies:
            self.studies = {self.mock_study}
            old_block_names = [x for x in self.iat_data.columns if x.endswith(suffices)]
            new_block_names = [f"{self.mock_study}_{x}" for x in old_block_names]
            self.iat_data.rename(columns=dict(zip(old_block_names, new_block_names)), inplace=True)
            
    def __repr__(self):
        return f"{self.__class__.__name__}('data_file', grouped_by={self.grouped_by}, " \
               f"trial_response_separator={self.trial_response_separator!r}, " \
               f"suffix_responses={self.suffix_responses!r}, " \
               f"suffix_trials={self.suffix_trials!r}, " \
               f"suffix_conditions={self.suffix_conditions!r}, " \
               f"congruency_labels={self.congruency_labels})"
        
    def _transpose_block_wide_to_long(self, suffix, separator, trial_data_name):
        """Transpose block data from the wide format to the long format, trial responses and stimuli
        :param suffix: str, determine the columns to be processed
        :param separator: str, the separator between consecutive trials
        :param trial_data_name: str, the trial data column name
        :return DataFrame, the transposed DataFrame
        :raise ValueError when found no columns for the responses data"""
        block_cols = [x for x in self.iat_data.columns if x.endswith(suffix)]
        if not block_cols:
            if suffix == self.suffix_trials:
                # it can happen when the researchers don't record the trial stimuli data
                return pd.DataFrame(
                    columns=[*reversed(self.grouped_by), "block_number", "trial_number", trial_data_name])
            else:
                raise ValueError("No columns were found for the responses data")
            
        block_wide = self.iat_data.melt(
            id_vars=self.grouped_by[1],
            value_vars=block_cols,
            value_name="block_data",
            var_name="block_counter"
        ).dropna()
        block_wide[self.grouped_by[0]] = block_wide["block_counter"].str[:-len(f"block1{suffix}") - 1]
        block_wide["block_number"] = block_wide["block_counter"].str[-len(suffix) - 1].map(int)
        trial_data = block_wide['block_data'].str.split(separator, expand=True)
        trial_columns = [x + 1 for x in trial_data.columns]
        block_wide[trial_columns] = trial_data
        block_long = block_wide.melt(
            id_vars=[*reversed(self.grouped_by), "block_number"],
            value_vars=trial_columns,
            value_name=trial_data_name,
            var_name="trial_number"
        ).dropna()
        return block_long
    
    def _transpose_block_conditions(self, study):
        """Create the block conditions by study"""
        condition_data = self.iat_data[[self.grouped_by[1], f"{study}_{self.suffix_conditions}"]].copy()
        study_conditions = condition_data[f"{study}_{self.suffix_conditions}"].str.split("|", expand=True)
        block_counters = [x + 1 for x in study_conditions.columns]
        condition_data[block_counters] = study_conditions
        condition_data[self.grouped_by[0]] = study
        return condition_data.melt(
            id_vars=reversed(self.grouped_by),
            value_vars=block_counters,
            value_name="block_condition",
            var_name="block_number"
        )
    
    def clean_up(self):
        """Clean up the IAT data response
        :return The cleaned up trial-level DataFrame
        :raise
            ValueError, when the response can't be casted
            AssertionError, when the trial number isn't the same from the trial number generated from its positioning
        """
        
        latency_data = self._transpose_block_wide_to_long(
            self.suffix_responses, self.trial_response_separator, "trial_response")
        stimulus_data = self._transpose_block_wide_to_long(
            self.suffix_trials, ",", "trial_stimulus")
        if not stimulus_data.empty:
            trial_data = latency_data.merge(stimulus_data,
                                            on=[*reversed(self.grouped_by), "block_number", "trial_number"])
        else:
            trial_data = latency_data
        
        def _score_trial_response(x):
            scored_data = [np.nan, np.nan, np.nan]
            if isinstance(x, str) and x != "None":
                try:
                    correct_index = max(x.find("Y"), x.find("N"))
                    if correct_index > 0:
                        scored_data = int(x[:correct_index]), x[correct_index], int(x[correct_index + 1:])
                except ValueError:
                    raise ValueError("can't cast the response")
            return pd.Series(scored_data)
        
        trial_cols = 'trial_counter trial_correct reaction_time'.split()
        trial_data[trial_cols] = trial_data['trial_response'].apply(_score_trial_response)
        trial_number_error_msg = "Split trial numbers are different from the trial number prefixes in the " \
                                 "block responses."
        assert pd.Series((trial_data["trial_number"] != trial_data["trial_counter"])).sum() == 0, trial_number_error_msg

        conditions = {label: x for x, labels in self.congruency_labels.items() for label in labels}
        block_conditions = pd.concat([self._transpose_block_conditions(study) for study in self.studies])
        block_conditions["task"] = block_conditions["block_condition"].map(
            lambda x: conditions.get(x, "sin")
        )
        block_conditions['task_block_counter'] = block_conditions.groupby([*reversed(self.grouped_by), "task"]).\
            cumcount().map(lambda x: x+1)
        
        trial_merged = trial_data.merge(block_conditions, on=[*reversed(self.grouped_by), "block_number"])
        
        self.iat_data_clean = trial_merged.drop(["trial_counter", "trial_response"], axis=1).sort_values(
            by=[*self.grouped_by, "block_number"]).dropna().reset_index(drop=True)
        return self.iat_data_clean


class IATAlgorithmName(Enum):
    """The list of supported algorithms"""
    CONVENTIONAL = "conventional"
    IMPROVED = "improved"
    

class IATErrorPenalty(Enum):
    """Error latency penalty approach"""
    ABSOLUTE = 0
    RELATIVE = 1
    BLOCK_MEAN = 2
    
    @property
    def description(self):
        """Description of the error penalty approach"""
        if self == self.ABSOLUTE:
            return "Block mean of correct responses + penalty time in ms"
        elif self == self.RELATIVE:
            return "Block mean of correct responses + penalty time in SD unit"
        else:
            return "Use latency to correct responses when correction is required after an error"


class IATAlgorithm:
    """Data model for the algorithm used in IAT data scoring"""
    def __init__(self, name, **params):
        """Initialize the instance for the algorithm
        :param name: str, the name of the algorithm
        :param params: dict, additional parameters for the algorithm
            When the algorithm is conventional, the following keyword parameters are supported:
                included_blocks: list, the blocks to be used for scoring
                rt_low_cutoff: int, float, the reaction time low cutoff
                rt_high_cutoff: int, float, the reaction time high cutoff
                recode_outliers: bool, whether outliers are recoded
                trials_to_drop: int, the number of trials at a block's beginning to drop
                allowed_error_rate: float, the percentage of error rate to drop subjects
                allowed_rt_upper: float, the minimum mean reaction time allowed to be included in the scoring
                
            When the algorithm is improved, the following keyword parameters are supported:
                included_blocks: list, the blocks to be used for scoring
                rt_low_cutoff: int, float, the reaction time low cutoff
                rt_high_cutoff: int, float, the reaction time high cutoff
                allowed_fast_rate: float, the percentage of fast responses, above which subjects are eliminated
                use_all_trials: bool, whether all trials are used after removing slow and fast responses
                rt_delete_cutoff: float, below which the trials are deleted, it's only valid when use_all_trials False
                pooled_sd_using_all: bool, whether to use all trials or only the correct trials to calculate SD
                replacement_option: IATErrorPenalty or 0, 1, 2, how to replace error trials
                rt_punishment: int or float, the penalty adjustment for error trials
            """
        self.name = name
        self.iat_data = None
        if name == IATAlgorithmName.CONVENTIONAL.name.lower():
            self.included_blocks = params.get("included_blocks", [4, 7])
            self.rt_low_cutoff = params.get("rt_low_cutoff", 300)
            self.rt_high_cutoff = params.get("rt_high_cutoff", 3000)
            self.recode_outliers = params.get("recode_outliers", True)
            self.trials_to_drop = params.get("trials_to_drop", 2)
            self.allowed_error_rate = params.get("allowed_error_rate", 0.1)
            self.allowed_rt_upper = params.get("allowed_rt_upper", 3000)
        elif name == IATAlgorithmName.IMPROVED.name.lower():
            self.included_blocks = params.get("included_blocks", [3, 4, 6, 7])
            self.rt_low_cutoff = params.get("rt_low_cutoff", 300)
            self.rt_high_cutoff = params.get("rt_high_cutoff", 10000)
            self.allowed_fast_rate = params.get("allowed_fast_rate", 0.1)
            self.use_all_trials = params.get("use_all_trials", True)
            self.rt_delete_cutoff = params.get("rt_delete_cutoff", 400)
            self.pooled_sd_using_all = params.get("pooled_sd_using_all", True)
            self.replacement_option = params.get("replacement_option", IATErrorPenalty.ABSOLUTE.value)
            self.rt_punishment = params.get("rt_punishment", 600)
            
    def __repr__(self):
        """Define the string representation of the instance"""
        params = self.__dict__.copy()
        del params["name"], params["iat_data"]
        params_str = ", ".join("=".join(map(str, item)) for item in params.items())
        return f"{self.__class__.__name__}({self.name!r}, {params_str})"
    
    def __str__(self):
        """Define the string when the instance is printed"""
        if self.name == IATAlgorithmName.CONVENTIONAL.name.lower():
            return f"""\nIncluded Blocks: {str(self.included_blocks)[1:-1]}\n
Reaction Time (Low Cutoff): {self.rt_low_cutoff} ms\n
Reaction Time (High Cutoff): {self.rt_high_cutoff} ms\n
Recode Outliers: {"Yes" if self.recode_outliers else "No"}\n
The First X Trials to Drop: {self.trials_to_drop}\n
Allowed Error Rate: {self.allowed_error_rate:.2%}\n
Allowed Latency Limit as Slow Responders: mean RT >{self.allowed_rt_upper} ms\n"""
        else:
            return f"""\nIncluded Blocks: {str(self.included_blocks)[1:-1]}\n
Reaction Time (Low Cutoff): {self.rt_low_cutoff} ms\n
Reaction Time (High Cutoff): {self.rt_high_cutoff} ms\n
Allowed Fast Rate: {self.allowed_fast_rate:.2%}\n
Use All Trials (if no, use the rt cutoff below to remove trials): {"Yes" if self.use_all_trials else "No"}\n
Reaction Time Cutoff (for deletion) = {self.rt_delete_cutoff}ms\n
Pooled SD: {"Using all trials" if self.pooled_sd_using_all else "Using correct trials only"}\n
Error Trial Replacement: {IATErrorPenalty(self.replacement_option).description}\n
Error Trial Penalty (ms or SD unit): {self.rt_punishment}"""
    
    def _calculate_reliability(self, trial_data, scored_iat_df=None):
        """Calculate the reliability of the IAT measure
        :param trial_data: DataFrame, the trial level data
        :param scored_iat_df: DataFrame, the scored IAT DataFrame, only applies for the improved algorithm
        """
        reliability_scores = list()
        
        def _compute_reliability_score(merged_df, col1, col2):
            def _spearman_brown_formula(x):
                return 2 * x / (1 + x)
    
            iat_corr = merged_df.groupby(self.iat_data.grouped_by[0])[[col1, col2]].corr().iloc[0::2, -1]
            iat_corr.index = iat_corr.index.droplevel(1)
            return iat_corr.map(_spearman_brown_formula)
        
        if self.name == IATAlgorithmName.CONVENTIONAL.name.lower():
            used_column = "iat_score_logged"
            _algorithm = self._process_data_conventional
        else:
            used_column = "iat_score"
            _algorithm = self._process_data_improved
            reliability_scores.append(_compute_reliability_score(scored_iat_df, "iat_score_1", "iat_score_2"))
        
        used_columns = [*self.iat_data.grouped_by, used_column]
        _, odd_scored_iat_df = _algorithm(trial_data[trial_data["trial_number"] % 2 == 1])
        _, even_scored_iat_df = _algorithm(trial_data[trial_data["trial_number"] % 2 == 0])
        odd_even_iat = odd_scored_iat_df[used_columns].merge(
            even_scored_iat_df[used_columns],
            on=self.iat_data.grouped_by,
            suffixes=("_odd", "_even")
        )
        reliability_scores.append(_compute_reliability_score(odd_even_iat, f"{used_column}_odd", f"{used_column}_even"))
        
        return reliability_scores
    
    def _apply_conventional(self):
        """Apply the conventional algorithm
        :return tuple, (DataFrame, DataFrame), the scored summary and response-level data"""
        trial_data = self.iat_data.iat_data_clean
        summary_df, scored_iat_df = self._process_data_conventional(trial_data)
        summary_df["reliability_by_odd_even"] = self._calculate_reliability(trial_data)
        return self._clean_up_scored_data(summary_df, scored_iat_df)
    
    def _process_data_shared(self, trial_data):
        iat_data_report = pd.DataFrame()
        grouped_by = self.iat_data.grouped_by
    
        # The total trial count by session
        iat_data_report["total_trial_count"] = trial_data.groupby(grouped_by).size()
        iat_data_report["total_error_trial_count"] = \
            trial_data[trial_data["trial_correct"] == "N"].groupby(grouped_by).size()
        iat_data_report["overall_error_rate"] = \
            iat_data_report["total_error_trial_count"] / iat_data_report["total_trial_count"]
        iat_data_report.fillna(0, inplace=True)
        return iat_data_report, grouped_by
    
    def _process_data_conventional(self, trial_data: pd.DataFrame):
        """Process data using the conventional algorithm
        :param trial_data: DataFrame, the trial-level data to be scored
        :return tuple, (DataFrame, DataFrame), the scored summary and response-level data"""
        iat_data_report, grouped_by = self._process_data_shared(trial_data)
        
        # Use the needed blocks and trials
        used_data = trial_data[trial_data['block_number'].isin(self.included_blocks) &
                               (trial_data['trial_number'] > self.trials_to_drop)].reset_index(drop=True)

        # The used trial count by session
        iat_data_report["used_trial_count"] = used_data.groupby(grouped_by).size()
        
        # Identify the error rates
        iat_data_report["error_trial_count"] = used_data[used_data['trial_correct'] == "N"].groupby(grouped_by).size()
        iat_data_report["error_trial_count"].fillna(0, inplace=True)
        iat_data_report["error_rate"] = iat_data_report["error_trial_count"] / iat_data_report["used_trial_count"]
        
        # Identify the average response time using all included trials
        iat_data_report["rt_mean"] = used_data.groupby(grouped_by)["reaction_time"].mean()
        
        included_responses = iat_data_report[
            (iat_data_report['error_rate'] < self.allowed_error_rate) &
            (iat_data_report['rt_mean'] < self.allowed_rt_upper)
            ].reset_index()[grouped_by]
        
        used_data = included_responses.merge(used_data, how="left", on=grouped_by)
        
        if self.recode_outliers:
            used_data['rt_recoded'] = used_data['reaction_time'].clip(
                lower=self.rt_low_cutoff,
                upper=self.rt_high_cutoff
            )
        else:
            used_data['rt_recoded'] = used_data['reaction_time'].map(
                lambda x: x if self.rt_low_cutoff <= x <= self.rt_high_cutoff else np.nan
            )
        used_data['rt_logged'] = np.log10(used_data['rt_recoded'])
        
        rt_mean_df = used_data.groupby([*grouped_by, 'task'], as_index=False).mean()
        calculated_iat = rt_mean_df.pivot(
            index=grouped_by,
            columns='task',
            values=['rt_recoded', 'rt_logged']
        ).reset_index()
        calculated_iat.columns = [x[0] + '_' + x[1] if x[1] else x[0] for x in calculated_iat.columns]
        
        calculated_iat['iat_score_raw'] = calculated_iat['rt_recoded_inc'] - calculated_iat['rt_recoded_con']
        calculated_iat['iat_score_logged'] = calculated_iat['rt_logged_inc'] - calculated_iat['rt_logged_con']
        
        scored_iat_df = iat_data_report.reset_index().merge(calculated_iat, how="left", on=grouped_by)
        summary_gb = scored_iat_df.groupby("study")
        summary_df = summary_gb.agg(
            trial_count_all_blocks=("total_trial_count", sum),
            trial_count_used_blocks=("used_trial_count", sum),
            trial_count_error=("error_trial_count", sum),
            pct_error_min_total=("overall_error_rate", np.min),
            pct_error_max_total=("overall_error_rate", np.max),
            pct_error_sd_total=("overall_error_rate", np.std),
            pct_error_mean_total=("overall_error_rate", np.mean),
            pct_error_mean_used_trials=("error_rate", np.mean),
            iat_raw_score_mean=("iat_score_raw", np.mean),
            iat_raw_score_min=("iat_score_raw", np.min),
            iat_raw_score_max=("iat_score_raw", np.max),
            iat_raw_score_sd=("iat_score_raw", np.std),
            rt_raw_con_mean=("rt_recoded_con", np.mean),
            rt_raw_con_min=("rt_recoded_con", np.min),
            rt_raw_con_max=("rt_recoded_con", np.max),
            rt_raw_con_sd=("rt_recoded_con", np.std),
            rt_raw_inc_mean=("rt_recoded_inc", np.mean),
            rt_raw_inc_min=("rt_recoded_inc", np.min),
            rt_raw_inc_max=("rt_recoded_inc", np.max),
            rt_raw_inc_sd=("rt_recoded_inc", np.std),
            iat_log_score_mean=("iat_score_logged", np.mean),
            iat_log_score_min=("iat_score_logged", np.min),
            iat_log_score_max=("iat_score_logged", np.max),
            iat_log_score_sd=("iat_score_logged", np.std),
            rt_log_con_mean=("rt_logged_con", np.mean),
            rt_log_con_min=("rt_logged_con", np.min),
            rt_log_con_max=("rt_logged_con", np.max),
            rt_log_con_sd=("rt_logged_con", np.std),
            rt_log_inc_mean=("rt_logged_inc", np.mean),
            rt_log_inc_min=("rt_logged_inc", np.min),
            rt_log_inc_max=("rt_logged_inc", np.max),
            rt_log_inc_sd=("rt_logged_inc", np.std)
        )
        return summary_df, scored_iat_df
    
    def _process_data_improved(self, trial_data: pd.DataFrame):
        """Process data using the improved algorithm
        :param trial_data: DataFrame, the trial-level data to be scored
        :return tuple, (DataFrame, DataFrame), the scored summary and response-level data"""
        iat_data_report, grouped_by = self._process_data_shared(trial_data)
        
        # Use the needed blocks
        used_data = trial_data[trial_data['block_number'].isin(self.included_blocks)].reset_index(drop=True)
    
        # The used trial count by session
        iat_data_report["used_trial_count"] = used_data.groupby(grouped_by).size()
    
        # Identify the high latency trials
        used_data["above_rt_upper_limit"] = used_data['reaction_time'] > self.rt_high_cutoff
        iat_data_report["high_latency_trial_count"] = \
            used_data[used_data["above_rt_upper_limit"]].groupby(grouped_by).size()
    
        # Identify the sessions with high fast trials
        iat_data_report["fast_trial_count"] = \
            used_data[used_data['reaction_time'] < self.rt_low_cutoff].groupby(grouped_by).size()
        iat_data_report["fast_trial_pct"] = iat_data_report["fast_trial_count"] / iat_data_report["used_trial_count"]
        too_many_fast_trial = pd.DataFrame((iat_data_report["fast_trial_pct"] > self.allowed_fast_rate).
                                           rename("too_many_fast_trial")).reset_index()
        used_data_merged = used_data.merge(too_many_fast_trial, on=grouped_by)
    
        # Identify the trials with fast responses that can be deleted if the option is selected
        used_data_merged["below_rt_fast_limit"] = used_data_merged['reaction_time'] < self.rt_delete_cutoff
        iat_data_report["fast_latency_trial_count"] = \
            used_data_merged[used_data_merged["below_rt_fast_limit"]].groupby(grouped_by).size()
    
        # Eliminate the trials with high latency and the sessions with high percentage of fast trials
        keep_condition = ~used_data_merged["above_rt_upper_limit"] & ~used_data_merged["too_many_fast_trial"]
    
        # Delete the fast trials if applicable
        if not self.use_all_trials:
            keep_condition = keep_condition & ~used_data_merged["below_rt_fast_limit"]
    
        used_data = used_data_merged[keep_condition].copy()
    
        iat_data_report["final_used_trial_count"] = used_data.groupby(grouped_by).size()
        iat_data_report["error_trial_count"] = used_data[used_data["trial_correct"] == "N"].groupby(grouped_by).size()
        iat_data_report["error_rate"] = iat_data_report["error_trial_count"] / iat_data_report["final_used_trial_count"]
        iat_data_report.fillna(0, inplace=True)
    
        def _make_na_values(x):
            if x["final_used_trial_count"] == 0:
                x["error_trial_count"] = x["error_rate"] = np.nan
            return pd.Series([x["error_trial_count"], x["error_rate"]])
    
        iat_data_report[["error_trial_count", "error_rate"]] = iat_data_report.apply(_make_na_values, axis=1)
    
        # Recode error latencies
        error_penalty = IATErrorPenalty(self.replacement_option)
        if error_penalty in (IATErrorPenalty.ABSOLUTE, IATErrorPenalty.RELATIVE):
            rt_block = used_data[used_data['trial_correct'] == "Y"]. \
                groupby([*grouped_by, 'block_number'])['reaction_time']. \
                mean().rename('rt_block_mean').reset_index()
            used_data = used_data.merge(rt_block, on=[*grouped_by, 'block_number'])
            if error_penalty == IATErrorPenalty.RELATIVE:
                std_block = used_data[used_data['trial_correct'] == "Y"]. \
                    groupby([*grouped_by, 'block_number'])['reaction_time']. \
                    std().rename('rt_block_std').reset_index()
                used_data = used_data.merge(std_block, on=[*grouped_by, 'block_number'])
        
            def _recode_error_latency(x):
                if x['trial_correct'] == 'Y':
                    return x['reaction_time']
                if error_penalty == IATErrorPenalty.ABSOLUTE:
                    return x['rt_block_mean'] + self.rt_punishment
                return x['rt_block_mean'] + self.rt_punishment * x['rt_block_std']
        
            used_data['rt_recoded'] = used_data.apply(_recode_error_latency, axis=1)
        else:
            used_data['rt_recoded'] = used_data['reaction_time']
    
        rt_task = used_data.groupby([*grouped_by, 'task', 'task_block_counter'], as_index=False)['rt_recoded'].mean()
        iat_scores_task = rt_task.pivot(
            index=[*grouped_by, 'task_block_counter'],
            values=['rt_recoded'],
            columns=['task']
        ).reset_index()
        iat_scores_task.columns = [x[0] + '_' + x[1] if x[1] else x[0] for x in iat_scores_task.columns]
    
        pooled_std = (used_data if self.pooled_sd_using_all else used_data[used_data['trial_correct'] == "Y"]). \
            groupby([*grouped_by, 'task_block_counter'])['reaction_time'].std().rename('pooled_std').reset_index()
        iat_scores = iat_scores_task.merge(pooled_std, on=[*grouped_by, 'task_block_counter'])
        iat_scores['iat_score'] = \
            (iat_scores['rt_recoded_inc'] - iat_scores['rt_recoded_con']) / iat_scores['pooled_std']
    
        iat_scores_session = iat_scores.pivot(
            index=grouped_by,
            values=["rt_recoded_con", "rt_recoded_inc", "pooled_std", "iat_score"],
            columns='task_block_counter'
        ).reset_index()
    
        iat_scores_session.columns = [f"{x[0]}_{x[1]}" if x[1] else x[0] for x in iat_scores_session.columns]
    
        def _calculate_means(x):
            iat_score_mean = np.mean(x[["iat_score_1", "iat_score_2"]])
            rt_recoded_con_mean = np.mean(x[["rt_recoded_con_1", "rt_recoded_con_2"]])
            rt_recoded_inc_mean = np.mean(x[["rt_recoded_inc_1", "rt_recoded_inc_2"]])
            return pd.Series([iat_score_mean, rt_recoded_con_mean, rt_recoded_inc_mean])
    
        iat_scores_session["iat_score rt_recoded_con rt_recoded_inc".split()] = \
            iat_scores_session.apply(_calculate_means, axis=1)
    
        scored_iat_df = iat_data_report.reset_index().merge(iat_scores_session, on=grouped_by, how="left")
    
        summary_gb = scored_iat_df.groupby("study")
        summary_df = summary_gb.agg(
            total_response_count=("final_used_trial_count", lambda x: (x > -1).sum()),
            used_response_count=("final_used_trial_count", lambda x: (x > 0).sum()),
            excluded_response_count=("final_used_trial_count", lambda x: (x < 1).sum()),
            trial_count_all_blocks=("total_trial_count", sum),
            trial_count_errors_all_blocks=("total_error_trial_count", sum),
            pct_error_min_total=("overall_error_rate", np.min),
            pct_error_max_total=("overall_error_rate", np.max),
            pct_error_sd_total=("overall_error_rate", np.std),
            pct_error_mean_total=("overall_error_rate", np.mean),
            trial_count_used_blocks=("used_trial_count", sum),
            trial_count_high_latency=("high_latency_trial_count", sum),
            trial_count_fast=("fast_trial_count", sum),
            pct_fast_mean=("fast_trial_pct", np.mean),
            trial_count_fast_latency=("fast_latency_trial_count", sum),
            trial_count_final=("final_used_trial_count", sum),
            trial_count_errors_used_trials=("error_trial_count", sum),
            pct_error_mean_used_trials=("error_rate", np.mean),
            iat_score_mean=("iat_score", np.mean),
            iat_score_min=("iat_score", np.min),
            iat_score_max=("iat_score", np.max),
            iat_score_sd=("iat_score", np.std),
            rt_con_mean=("rt_recoded_con", np.mean),
            rt_con_min=("rt_recoded_con", np.min),
            rt_con_max=("rt_recoded_con", np.max),
            rt_con_sd=("rt_recoded_con", np.std),
            rt_inc_mean=("rt_recoded_inc", np.mean),
            rt_inc_min=("rt_recoded_inc", np.min),
            rt_inc_max=("rt_recoded_inc", np.max),
            rt_inc_sd=("rt_recoded_inc", np.std),
        )
        return summary_df, scored_iat_df
    
    def _apply_improved(self):
        """Apply the improved algorithm
        :return tuple, (DataFrame, DataFrame), the scored summary and response-level data"""
        trial_data = self.iat_data.iat_data_clean
        summary_df, scored_iat_df = self._process_data_improved(trial_data)
        summary_df["reliability_by_task"], summary_df["reliability_by_odd_even"] = \
            tuple(self._calculate_reliability(trial_data, scored_iat_df))
        return self._clean_up_scored_data(summary_df, scored_iat_df)
    
    def _clean_up_scored_data(self, summary_df, scored_iat_df):
        """Clean up the return data set
        :return tuple, (DataFrame, DataFrame), the scored summary and response-level data"""
        overall_summary_df = summary_df.reset_index()
        if len(self.iat_data.studies) == 1 and self.iat_data.studies.pop() == IATData.mock_study:
            scored_iat_df.drop(columns=self.iat_data.grouped_by[0], inplace=True)
            overall_summary_df.drop(columns=self.iat_data.grouped_by[0], inplace=True)
        return overall_summary_df, scored_iat_df
    
    def process_data(self, iat_data: IATData):
        """Process the data using the current algorithm
        :param iat_data: IATData, the IATData instance
        :return tuple, (DataFrame, DataFrame), the scored summary and response-level data"""
        self.iat_data = iat_data
        if self.name == IATAlgorithmName.CONVENTIONAL.value:
            return self._apply_conventional()
        elif self.name == IATAlgorithmName.IMPROVED.value:
            return self._apply_improved()



import base64
import json
import streamlit as st
import qualtrics_tools
import iat_scorer
import script_generator
import web_utils

IATTask = script_generator.IATTask
SessionState = web_utils.SessionState
sidebar = st.sidebar
session_state = SessionState.get(
    templates=IATTask.templates(),
    working_task=None,
    iat_data=None,
    tool=qualtrics_tools.QualtricsTool()
)
tool = session_state.tool


def _set_width(width):
    st.markdown(
        f"""<style>.reportview-container .main .block-container{{max-width: {width}px;}}</style>""",
        unsafe_allow_html=True,
    )


def _preview_instructions(instruction, container):
    evaluated_instruction = eval(instruction)
    if isinstance(evaluated_instruction, list) and evaluated_instruction:
        for page_i, page in enumerate(evaluated_instruction, 1):
            container.markdown(f"__Page {page_i}__")
            container.markdown(page, unsafe_allow_html=True)
            container.markdown("___")


def create_downloadable_link(encoded, file_name, link_name, container, /, requires_encoding=True, filetype='txt'):
    if requires_encoding:
        encoded = encoded.encode()
    encoded_script = base64.b64encode(encoded).decode()
    href = f'<a href="data:file/{filetype};base64,{encoded_script}" ' \
           f'download="{file_name}">{link_name}</a>'
    container.markdown(href, unsafe_allow_html=True)


def _load_generator():
    sidebar.markdown("____")
    sidebar.header("Templates")

    templates = session_state.templates
    selected_template_name = sidebar.radio("Choose Template", IATTask.templates_names)
    selected_template_index = IATTask.templates_names.index(selected_template_name)
    if sidebar.button("Reload Selected Template"):
        templates[selected_template_index] = IATTask.reset_template(selected_template_name)
    working_task = session_state.working_task = templates[selected_template_index]
    sidebar.markdown(
        "Reloading template will overwrite your progress by resetting the configurations to the initial status."
    )

    sidebar.markdown("____")
    sidebar.header("Actions")
    sidebar.markdown("#### Generate Qualtrics File (QSF)")
    sidebar.markdown("Use this file to create a new survey. "
                     "If your Qualtrics subscription doesn't support this feature, "
                     "use the _Generate Script_ button below instead.")
    qsf_button = sidebar.button("Generate Template (QSF)")
    if qsf_button:
        template_js = working_task.generate_template_file()
        create_downloadable_link(
            template_js,
            "iat_qualtrics_template.qsf",
            "Download Qualtrics Template File",
            sidebar
        )

    sidebar.markdown("#### Generate JavaScript Script (JS)")
    sidebar.markdown("You'll build your survey yourself. The script will be added to the presentation question.")
    js_button = sidebar.button("Generate Script (JS)")
    if js_button:
        generated_script = working_task.generate_script()
        create_downloadable_link(
            generated_script,
            "iat_questionjs_script.js",
            "Download Question JS Code",
            sidebar
        )

    st.header("IAT Qualtrics Survey Generator")
    st.subheader("")
    st.subheader("Target Concepts")
    shared_stimuli_instruction = "(words or URLs to images), separated by commas. Enclose each word or URL with " \
                                 "single or double quotes and your entire list with square brackets for coding " \
                                 "purposes."
    target_positive_col, target_negative_col = st.beta_columns(2)

    working_task.target_positive_concept = target_positive_col.text_input(
        "Positive Target Concept",
        working_task.target_positive_concept
    )
    working_task.target_positive_stimuli = target_positive_col.text_area(
        f"Positive Target Stimuli {shared_stimuli_instruction}",
        working_task.target_positive_stimuli
    )
    working_task.target_negative_concept = target_negative_col.text_input(
        "Negative Target Concept",
        working_task.target_negative_concept
    )
    working_task.target_negative_stimuli = target_negative_col.text_area(
        f"Negative Target Stimuli {shared_stimuli_instruction}",
        working_task.target_negative_stimuli
    )

    stimulus_references = ["words", "pictures"]
    stimulus_medias = ["text", "image"]

    working_task.target_stimulus_reference = st.selectbox(
        "Target Stimuli Media (for references used in instructions)",
        stimulus_references,
        stimulus_references.index(working_task.target_stimulus_reference)
    )
    working_task.target_stimulus_media = st.selectbox(
        "Target Stimuli Media (for configuring web elements used in presentations). In some cases, for "
        "presentation consistency, words can be prepared as images.",
        stimulus_medias,
        stimulus_medias.index(working_task.target_stimulus_media)
    )

    st.subheader("Attribute Concepts")
    attribute_positive_col, attribute_negative_col = st.beta_columns(2)
    working_task.attribute_positive_concept = attribute_positive_col.text_input(
        "Positive Attribute Concept",
        working_task.attribute_positive_concept
    )
    working_task.attribute_positive_stimuli = attribute_positive_col.text_area(
        f"Positive Attribute Stimuli {shared_stimuli_instruction}",
        working_task.attribute_positive_stimuli
    )
    working_task.attribute_negative_concept = attribute_negative_col.text_input(
        "Negative Attribute Concept",
        working_task.attribute_negative_concept
    )
    working_task.attribute_negative_stimuli = attribute_negative_col.text_area(
        f"Negative Attribute Stimuli {shared_stimuli_instruction}",
        working_task.attribute_negative_stimuli
    )

    working_task.attribute_stimulus_reference = st.selectbox(
        "Attribute Stimuli Media (for references used in instructions)",
        stimulus_references,
        stimulus_references.index(working_task.attribute_stimulus_reference)
    )
    working_task.attribute_stimulus_media = st.selectbox(
        "Attribute Stimuli Media (for configuring web elements used in presentations). In some cases, for"
        "presentation consistency, words can be prepared as images.",
        stimulus_medias,
        stimulus_medias.index(working_task.attribute_stimulus_media)
    )
    st.markdown("____")

    st.header("Advanced Configurations")

    st.subheader("Study Name")
    working_task.study_name = st.text_input(
        "The study name will be useful if you have multiple IATs in one survey.",
        working_task.study_name
    )
    if working_task.study_name:
        st.markdown(
            f"The block responses will be saved as several embedded data with the name using the format: "
            f"{working_task.study_name}_block1Responses, {working_task.study_name}_block2Responses..."
            f"Setting the study name is necessary if you have multiple IAT experiments in a Qualtrics survey."
        )
    else:
        st.markdown(
            f"The block responses will be saved as several embedded data with the name using the format: "
            f"block1Responses, block2Responses..."
        )
    st.markdown("___")

    st.subheader("Switch Attributes or Targets for Block 5")
    switch_options = ["Attribute", "Target"]
    working_task.switch_attributes = st.selectbox(
        "Indicate whether you want to switch attributes or targets.",
        switch_options
    ) == switch_options[0]
    if working_task.switch_attributes:
        switch_text = "Block 5 will be the switched attributes opposite to Block 2."
    else:
        switch_text = "Block 5 will be the switched targets opposite to Block 1."
    st.markdown(switch_text)
    st.markdown("___")

    st.subheader("Estimated Duration")
    working_task.estimated_duration = st.text_input(
        "Estimated Duration for Completing Survey (please include the unit for the time)",
        working_task.estimated_duration
    )
    st.markdown(
        f"In the instruction, your subjects will be notified that it will take them about "
        f"{working_task.estimated_duration} for completing the study."
    )
    st.markdown("___")

    st.subheader("Counter-Balancing of Block Conditions")
    working_task.counter_balancing = st.checkbox("Enable Counter-Balancing", True)
    if working_task.counter_balancing:
        st.markdown("The order of congruency and left vs. right will be counter-balanced between subjects.")
    else:
        st.markdown("WARNING: The order of the blocks will be fixed for all subjects: congruent first "
                    "then incongruent.")
    st.markdown("___")

    st.subheader("Correction Requirement")
    working_task.requires_correction = st.checkbox(
        "Requires Correction After Making a Mistake",
        working_task.requires_correction
    )
    if working_task.requires_correction:
        st.markdown("Your task requires that the subject correct the error responses before it proceeds.")
        rt_options = {"After the first error response": "error", "After the correct response": "correct"}
        working_task.error_rt_option = rt_options[
            st.selectbox("How to Record Reaction Time for Error Responses", list(rt_options.keys()), 0)
        ]
        if working_task.error_rt_option == list(rt_options.values())[0]:
            st.markdown("For error responses, the reaction time will be recorded upon the first error attempt.")
        else:
            st.markdown("For error responses, the reaction time will be recorded upon the correct attempt.")
    else:
        working_task.auto_advance_delay = st.number_input("The delay in ms before the task proceeds "
                                                          "after a wrong responses", 300.0)
        st.markdown(f"Your task will automatically proceed {working_task.auto_advance_delay} ms after error responses.")
    st.markdown("___")

    st.subheader("Response Keys")
    st.markdown(
        "You can find the list of key codes from "
        "[Mozilla Website](https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent/code/code_values)."
    )
    key_column_titles = ["Code (used for recording responses)", "Name (used for instructions)"]
    key_row_titles = ["Left Key", "Right Key", "Advance Key (to launch blocks)"]
    response_key_code_col, response_key_name_col = st.beta_columns(len(key_column_titles))
    response_key_code_col.markdown(f"__{key_column_titles[0]}__")
    response_key_name_col.write(f"__{key_column_titles[1]}__")
    working_task.left_key_code = response_key_code_col.text_input(
        key_row_titles[0],
        working_task.left_key_code,
        key=key_row_titles[0] + key_column_titles[0]
    )
    working_task.right_key_code = response_key_code_col.text_input(
        key_row_titles[1],
        working_task.right_key_code,
        key=key_row_titles[1] + key_column_titles[0]
    )
    working_task.advance_key_code = response_key_code_col.text_input(
        key_row_titles[2],
        working_task.advance_key_code,
        key=key_row_titles[2] + key_column_titles[0]
    )
    working_task.left_key_name = response_key_name_col.text_input(
        key_row_titles[0],
        working_task.left_key_name,
        key=key_row_titles[0] + key_column_titles[1]
    )
    working_task.right_key_name = response_key_name_col.text_input(
        key_row_titles[1],
        working_task.right_key_name,
        key=key_row_titles[1] + key_column_titles[1]
    )
    working_task.advance_key_name = response_key_name_col.text_input(
        key_row_titles[2],
        working_task.advance_key_name,
        key=key_row_titles[2] + key_column_titles[1]
    )
    st.markdown(
        f"When your task is completed on a desktop computer, the subject will use __{working_task.left_key_code}__ and "
        f"__{working_task.right_key_code}__, namely _{working_task.left_key_name}_ and "
        f"_{working_task.right_key_name}_ for "
        f"left and right responses. The __{working_task.advance_key_code}__ with "
        f"the name _{working_task.advance_key_name}_ "
        f"will be used to advance the task."
    )
    st.markdown("___")

    st.subheader("Overall Instruction")
    st.write("Instructions should be prepared as HTTP texts, and they will be rendered page by page. Each page will be "
             "a text string. Please use square brackets to wrap these individual pages, separated by commas. Previews "
             "are shown on the right side.")
    st.write("__Instruction for Desktop Users__")
    instruction_desktop_text_col, instruction_desktop_preview_col = st.beta_columns(2)
    working_task.overall_instruction_desktop = instruction_desktop_text_col.text_area(
        "HTTP Text",
        working_task.get_overall_instruction(False),
        key="overall_instruction_desktop",
        height=500
    )
    _preview_instructions(working_task.overall_instruction_desktop, instruction_desktop_preview_col)

    st.write("__Instruction for Mobile Users__")
    instruction_mobile_text_col, instruction_mobile_preview_col = st.beta_columns(2)
    working_task.overall_instruction_mobile = instruction_mobile_text_col.text_area(
        "HTTP Text",
        working_task.get_overall_instruction(True),
        key="overall_instruction_mobile",
        height=700
    )
    _preview_instructions(working_task.overall_instruction_mobile, instruction_mobile_preview_col)
    st.markdown("___")

    st.subheader("Display of Stimuli as Examples")
    working_task.shows_examples = st.checkbox("Shows Examples")
    if working_task.shows_examples:
        st.markdown(
            "Your subjects will review the stimuli to know what stimuli belong to what categories before they start "
            "the classification portion."
        )
        working_task.example_duration = st.number_input(
            "Example Duration in Milliseconds",
            working_task.example_duration
        )
        st.markdown("__Instructions Before Showing Examples__ (The rules for preparing the instructions are the "
                    "same as the ones for the overall instructions.)")
        example_instruction_text_col, example_instruction_preview_col = st.beta_columns(2)
        working_task.example_instruction = example_instruction_text_col.text_area(
            "HTTP Text",
            working_task.get_example_instruction(),
            key="example_instruction",
            height=600
        )
        _preview_instructions(working_task.example_instruction, example_instruction_preview_col)
    else:
        st.markdown(
            "Your subjects won't review the stimuli used in the study as examples before they start "
            "the classification portion."
        )
    st.markdown("___")

    st.subheader("Reminder Instructions")
    st.write("These instructions will be displayed right before the classification portion, as a friendly reminder.")
    st.write("__Reminder Instruction for Desktop Users__")
    reminder_instruction_desktop_text_col, reminder_instruction_desktop_preview_col = st.beta_columns(2)
    working_task.reminder_instruction_desktop = reminder_instruction_desktop_text_col.text_area(
        "HTTP Text",
        working_task.get_reminder_instruction(False),
        key="reminder_instruction_desktop",
        height=600
    )
    _preview_instructions(working_task.reminder_instruction_desktop, reminder_instruction_desktop_preview_col)

    st.write("__Reminder Instruction for Mobile Users__")
    reminder_instruction_mobile_text_col, reminder_instruction_mobile_preview_col = st.beta_columns(2)
    working_task.reminder_instruction_mobile = reminder_instruction_mobile_text_col.text_area(
        "HTTP Text",
        working_task.get_reminder_instruction(True),
        key="reminder_instruction_mobile",
        height=600
    )
    _preview_instructions(working_task.reminder_instruction_mobile, reminder_instruction_mobile_preview_col)
    st.markdown("___")

    st.subheader("Colors")
    color_column_titles = [
        "Attribute Word Stimulus Color",
        "Attribute Label Color",
        "Target Word Stimulus Color",
        "Target Label Color"
    ]
    color_columns = st.beta_columns(len(color_column_titles))
    working_task.attribute_word_color = color_columns[0].color_picker(
        color_column_titles[0],
        working_task.attribute_word_color
    )
    working_task.attribute_label_color = color_columns[1].color_picker(
        color_column_titles[1],
        working_task.attribute_label_color
    )
    working_task.target_word_color = color_columns[2].color_picker(
        color_column_titles[2],
        working_task.target_word_color
    )
    working_task.target_label_color = color_columns[3].color_picker(
        color_column_titles[3],
        working_task.target_label_color
    )
    st.markdown("___")

    st.subheader("Inter-Trial Interval (ITI)")
    working_task.inter_trial_interval = st.number_input(
        "ITI in Milliseconds",
        working_task.inter_trial_interval
    )
    st.markdown(f"Consecutive IAT trials will be separated by {working_task.inter_trial_interval} milliseconds.")
    st.markdown("___")

    st.subheader("Random Responses")
    working_task.prevent_random = st.checkbox("Prevent Random Responses", True)
    if working_task.prevent_random:
        working_task.minimum_allowed_reaction_time = st.number_input(
            "Minimum Allowed Reaction Time in Milliseconds",
            working_task.minimum_allowed_reaction_time
        )
        st.markdown(
            f"Responses shorter than {working_task.minimum_allowed_reaction_time} ms are considered random entries."
        )
        working_task.too_fast_response_error_message = st.text_input(
            "Warning Message for Random Responses",
            working_task.too_fast_response_error_message
        )
        st.markdown(
            f"When a random entry is identified, the screen will display: "
            f"{working_task.too_fast_response_error_message}"
        )
    else:
        st.markdown("The task won't detect whether subjects' responses are reasonably fast.")
    st.markdown("___")

    st.subheader("Minimum Preload Image Percent")
    working_task.minimum_preload_image_percent = st.slider(
        "The Percentage of Images Preloaded",
        min_value=0,
        max_value=100,
        value=working_task.minimum_preload_image_percent
    )
    st.markdown(
        "If your task uses images as stimuli, it's a good idea to preload images to increase the responsiveness of "
        "your task. If you don't want any preloading, set it to 0."
    )
    st.markdown("___")

    st.subheader("Inter-Trial Response Separator")
    working_task.inter_trial_response_separator = st.text_input(
        "The Delimiter",
        working_task.inter_trial_response_separator
    )
    st.markdown(
        f"Each block's responses will be saved as an embedded field in Qualtrics responses. In the block's "
        f"response, the trials' responses will be separated by {working_task.inter_trial_response_separator}. "
        f"Please don't use numbers or Y or N, which are part of the trial responses."
    )
    st.markdown("___")

    st.subheader("Block Specifications")
    st.write(
        "By default, the generator only works with the classic seven-block IAT. If you want to customize this "
        "behavior, you need to set the correct number of blocks, and specify the trial numbers and condition "
        "for each block."
    )
    working_task.block_trial_numbers = eval(st.text_input(
        "The number of trials for the blocks",
        working_task.block_trial_numbers
    ))
    st.markdown(
        f"The number of blocks for your task: {len(working_task.block_trial_numbers)}."
    )

    st.subheader("Automatic Testing")
    st.write(
        "When you specify a value that is greater than the minimally allowed response interval, "
        "the program can automatically enter the correct responses. When the value is "
        "smaller than or equal to the minimally allowed response interval, the program won't proceed by itself."
    )
    working_task.automatic_responses_delay = float(st.text_input(
        "Delay of automatic responses in milliseconds",
        working_task.automatic_responses_delay
    ))
    if working_task.automatic_responses_delay <= working_task.minimum_allowed_reaction_time:
        st.markdown(f"No automatic responses will be applied.")
    else:
        st.markdown(
            f"Automatic responses will be applied after the delay (ms): {working_task.automatic_responses_delay}"
        )


def _load_qualtrics_tools():
    sidebar.markdown("____")
    sidebar.header("Required API Configurations")
    sidebar.markdown(
        "To use the tools, you'll need to provide the necessary parameters in the first section. You can fetch "
        "these parameters from your Qualtrics account settings (Account Settings -> Qualtrics IDs). "
        "For more information about using these APIs, please "
        "refer to Qualtrics official website [Qualtrics API Quick Start]"
        "(https://api.qualtrics.com/instructions/docs/Instructions/Quick%20Start/qualtrics-api-quick-start.md)."
    )
    tool.api_token = sidebar.text_input("API Token", "")
    tool.data_center = sidebar.text_input("Data Center", "")
    tool.brand_center = sidebar.text_input("Brand Center", "")
    
    sidebar.markdown(
        "Your institute's designated Qualtrics website. For instance, MD Anderson Cancer Center uses "
        "'mdanderson.co1.qualtrics.com', in which case, you enter: mdanderson.co1 in the text field."
    )

    st.header("Qualtrics Tools")

    st.markdown("#### Upload Images to Qualtrics Graphics Library")
    st.markdown("Use this tool to upload images from your local computer to your Qualtrics Graphics Library.")
    upload_section = st.beta_expander("Upload Images")
    upload_section.markdown(
        "You can upload images from your computer. You can find the library ID # in your account settings "
        "(Account Settings -> Qualtrics IDs -> Libraries Table -> Row 1, the one starting with UR_)."
    )
    library_id = upload_section.text_input("Library ID #")
    qualtrics_folder = upload_section.text_input("Graphics Folder Name", "IAT Web App Upload")
    full_url = upload_section.checkbox("Generate Full URL", True)
    if full_url:
        url_note = "The full URLs for the images wil be generated. Please note that these images will be public, " \
                   "and anyone who has these URLs can access them."
    else:
        url_note = "Only the Qualtrics IDs of these images will be generated."
    upload_section.markdown(url_note)
    image_type = upload_section.selectbox("Image File Type", ['png', 'jpg', 'gif'])
    image_files = upload_section.file_uploader("Choose Images", ['png', 'jpg', 'gif'], True)
    upload_button = upload_section.button("Upload")
    if upload_button and image_files:
        image_urls = tool.upload_images_web(
            image_files,
            library_id,
            full_url,
            qualtrics_folder,
            image_type
        )
        upload_section.text_area("Image URLs", image_urls)
    st.markdown("____")

    st.markdown("#### Create Surveys")
    st.markdown("Use this tool to create surveys by setting JSON text or uploading QSF file.")
    create_section = st.beta_expander("Create Surveys")
    create_section.markdown(
        "You can create a survey either by specifying the template in a "
        "JSON format or by uploading the template file (.QSF). "
    )
    json_sources = ["Specify JSON Text", "Upload Qualtrics Survey File (.QSF)"]
    selected_json_source = create_section.selectbox("Required JSON Source", json_sources, 1)
    if selected_json_source == json_sources[0]:
        json_text = create_section.text_area("Enter the JSON Text")
        if json_text:
            json_text = json.loads(json_text)
    else:
        json_file = create_section.file_uploader("Choose Qualtrics Survey File", ['qsf'])
        if json_file:
            json_text = json.load(json_file)
    create_button = create_section.button("Create")
    if create_button and json_text:
        created_survey_id = tool.create_survey(json_text)
        created_survey_link = f"{tool.base_url}/Q/EditSection/Blocks?ContextSurveyID={created_survey_id}"
        create_section.markdown(f"Survey Link: [Qualtrics Survey (#{created_survey_id})]({created_survey_link})")
    st.markdown("____")

    st.markdown("#### Export Survey Responses")
    st.markdown("Use this tool to export survey responses for further processing.")
    export_section = st.beta_expander("Export Survey Responses")
    survey_id = export_section.text_input("Survey ID #")
    file_formats = ["csv", "tsv", "spss"]
    file_format = export_section.selectbox("File Format", file_formats)
    export_button = export_section.button("Export")
    if export_button:
        export_content = tool.export_responses(survey_id, file_format)
        create_downloadable_link(
            export_content,
            "survey_responses.zip",
            "Download survey_responses.zip",
            export_section,
            filetype="js",
            requires_encoding=False
        )
    st.markdown("____")

    st.markdown("#### Delete Images")
    st.markdown("Use this tool to delete images from your Qualtrics Graphics Library.")
    delete_section = st.beta_expander("Delete Images")
    delete_section.markdown("You can specify a list of either URLs or image ids.")
    delete_library_id = delete_section.text_input("Library ID #", key="for_deletion")
    delete_image_ids = delete_section.text_area("Image IDs")
    delete_button = delete_section.button("Delete")
    if delete_button and delete_image_ids:
        delete_image_ids = eval(delete_image_ids)
        if isinstance(delete_image_ids, list):
            delete_report = tool.delete_images(delete_library_id, delete_image_ids)
            delete_section.write(delete_report)
        else:
            delete_section.error("Please specify the list of image IDs.")
    st.markdown("____")

    st.markdown("#### Delete Survey")
    st.markdown("Use this tool to delete surveys from your Qualtrics Library.")
    delete_survey_section = st.beta_expander("Delete Survey")
    delete_survey_section.markdown("You can specify the survey id for deletion.")
    delete_survey_id = delete_survey_section.text_input("Survey ID")
    delete_confirm_text = st.text_input("Please enter Delete Survey to confirm your deletion.")
    delete_survey_button = delete_survey_section.button("Delete", key="delete_survey")
    if delete_survey_button and delete_survey_id and delete_confirm_text == "Delete Survey":
        delete_survey_report = tool.delete_survey(delete_survey_id)
        delete_survey_section.write(delete_survey_report)


def _load_scorer():
    sidebar.markdown("____")
    sidebar.header("Overview of Algorithms")
    sidebar.markdown(
        "__Algorithm Descriptions__ (Greenwald et al. Understanding and Using the Implicit Association Test: "
        "I. An Improved Scoring Algorithm. Journal of Personality and Social Psychology 2003 (85)2:192-216)"
    )
    sidebar.markdown("#### Conventional Algorithm")
    sidebar.markdown("1. Use data from B4 & B7 (counter-balanced order will be taken care of in the calculation).")
    sidebar.markdown(
        "2. Nonsystematic elimination of subjects for excessively slow responding and/or high error rates."
    )
    sidebar.markdown("3. Drop the first two trials of each block.")
    sidebar.markdown("4. Recode latencies outside 300/3,000 boundaries to the nearer boundary value.")
    sidebar.markdown("5. Log-transform the resulting values.")
    sidebar.markdown("6. Average the resulting values for each of the two blocks.")
    sidebar.markdown("7. Compute the difference: B7 - B4.")

    sidebar.markdown("#### Improved Algorithm")
    sidebar.markdown(
        "1. Use data from B3, B4, B6, & B7 (counter-balanced order will be taken care of in the calculation)."
    )
    sidebar.markdown(
        "2. Eliminate trials with latencies > 10,000 ms; Eliminate subjects for whom more than "
        "10% of trials have latency less than 300 ms."
    )
    sidebar.markdown("3. Use all trials; Delete trials with latencies below 400 ms (alternative).")
    sidebar.markdown(
        "4. Compute mean of correct latencies for each block. "
        "Compute SD of correct latencies for each block (alternative)."
    )
    sidebar.markdown(
        "5. Compute one pooled SD for all trials in B3 & B6, another for B4 & B7; Compute one pooled SD "
        "for correct trials in B3 & B6, another for B4 & B7 (alternative)."
    )
    sidebar.markdown(
        "6. Replace each error latency with block mean (computed in Step 5) + 600 ms; Replace each error "
        "latency with block mean + 2 x block SD of correct responses (alternative 1); "
        "Use latencies to correct responses when correction to error responses is required (alternative 2)."
    )
    sidebar.markdown("7. Average the resulting values for each of the four blocks.")
    sidebar.markdown("8. Compute two differences: B6 - B3 and B7 - B4.")
    sidebar.markdown("9. Divide each difference by its associated pooled-trials SD.")
    sidebar.markdown("10. Average the two quotients.")

    st.header("IAT Data Scorer")
    st.markdown("This scorer scores the data of the Qualtrics IAT survey in the CSV format.")
    data_file = st.file_uploader("IAT Data File", ["csv"])
    if data_file:
        session_state.iat_data = iat_scorer.IATData(data_file)
        session_state.iat_data.iat_data
        iat_data_clean = session_state.iat_data.clean_up()
        st.write(iat_data_clean)
        st.write(
            "Note: task (sin=single-label blocks, con=congruent blocks, inc=incongruent blocks. "
            "If you counter-balance the block order, it will be considered."
        )
        create_downloadable_link(
            iat_data_clean.to_csv(index=False),
            "IAT_Trial_Data.csv",
            "Download IAT Trial Data",
            st
        )
    iat_data = session_state.iat_data

    algorithms = ["Conventional Algorithm", "Improved Algorithm"]
    selected_algorithm_index = algorithms.index(st.selectbox("Choose Algorithm", algorithms, 1))
    calculation_params = dict()
    if selected_algorithm_index == 0:
        rt_cols = st.beta_columns(2)
        calculation_params['rt_low_cutoff'] = rt_cols[0].number_input(
            "Reaction Time (ms) Low Cutoff",
            value=300.0
        )
        calculation_params['rt_high_cutoff'] = rt_cols[1].number_input(
            "Reaction Time (ms) High Cutoff",
            value=3000.0
        )
        calculation_params['recode_outliers'] = st.checkbox(
            "Recode Outliers to Boundary Values",
            True
        )
        if calculation_params['recode_outliers']:
            st.write(
                f"Recode latencies outside "
                f"{calculation_params['rt_low_cutoff']}/{calculation_params['rt_high_cutoff']} "
                f"boundaries to the nearer boundary value."
            )
        else:
            st.write(
                f"Drop latencies outside "
                f"{calculation_params['rt_low_cutoff']}/{calculation_params['rt_high_cutoff']}"
            )
        calculation_params['trials_to_drop'] = st.number_input(
            "The Number of Trials to Drop for Each Block",
            value=2
        )
        st.write(f"The number of first trials to drop for each block: {calculation_params['trials_to_drop']}")
        calculation_params['allowed_error_rate'] = st.slider(
            "The Maximum Allowed Error Rate (%)",
            min_value=0.0,
            max_value=100.0,
            value=10.0
        ) / 100.0
        st.write(f"Eliminate subjects whose error rate is higher than {calculation_params['allowed_error_rate']:.2%}.")
        calculation_params['allowed_rt_upper'] = st.number_input(
            "Slow Responding Cutoff",
            value=3000.0
        )
        st.write(f"Eliminate subjects whose mean reaction time is "
                 f"longer than {calculation_params['allowed_rt_upper']} ms.")
    else:
        calculation_params['rt_high_cutoff'] = st.number_input(
            "Reaction Time (ms) High Cutoff",
            value=10000.0
        )
        st.write(f"Eliminate trials with latencies > {calculation_params['rt_high_cutoff']} ms")
        rt_cols = st.beta_columns(2)
        calculation_params['rt_low_cutoff'] = rt_cols[0].number_input(
            "Reaction Time (ms) Low Cutoff",
            value=300.0
        )
        calculation_params['allowed_fast_rate'] = rt_cols[1].slider(
            "Exclusion cutoff for subjects with > x% of too fast responses (reaction time < low cutoff)",
            min_value=0.0,
            max_value=100.0,
            value=10.0
        ) / 100.0
        st.write(f"Eliminate subjects for who more than {calculation_params['allowed_fast_rate']:%} of the "
                 f"trials have less than {calculation_params['rt_low_cutoff']} ms")

        use_trials_cols = st.beta_columns(2)
        calculation_params['use_all_trials'] = use_trials_cols[0].checkbox("Use all trials", True)
        if calculation_params['use_all_trials']:
            st.write("Use all trials after the first two steps")
        else:
            calculation_params['rt_delete_cutoff'] = use_trials_cols[1].number_input(
                "Delete Trials Whose Latency is Below (ms)",
                value=400.0
            )
            st.write(f"Delete trials with latencies below {calculation_params['rt_delete_cutoff']} ms.")

        pooled_sd_options = ["All Trials", "Correct Trials Only"]
        calculation_params['pooled_sd_using_all'] = st.selectbox(
            "Calculation of Pooled SD",
            pooled_sd_options
        ) == pooled_sd_options[0]
        if calculation_params['pooled_sd_using_all']:
            st.write("Use all trials in B3 and B6 to calculate the pooled SD. Use all trials in B4 and B7 to "
                     "calculate the pooled SD.")
        else:
            st.write("Use only correct trials in B3 and B6 to calculate the pooled SD. Use only correct trials in "
                     "B4 and B7 to calculate the pooled SD.")

        replacement_options = [
            "Block mean of correct responses + punishment time in ms",
            "Block mean of correct responses + punishment time in block SD of correct responses",
            "Use latency to correct responses when correction is required after an error"
        ]
        calculation_params['replacement_option'] = replacement_options.index(
            st.selectbox("Replacement of error latencies", replacement_options)
        )
        if calculation_params['replacement_option'] == 0:
            calculation_params['rt_punishment'] = st.slider(
                "Reaction Time (ms) Punishment for Error Trials",
                value=600,
                min_value=0,
                max_value=2000
            )
            st.write(f"Replace each error latency with block mean (computed in Step 5) + "
                     f"{calculation_params['rt_punishment']} ms")
        elif calculation_params['replacement_option'] == 1:
            calculation_params['rt_punishment'] = st.slider(
                "Reaction Time (ms) Punishment for Error Trials",
                value=2,
                min_value=0,
                max_value=10
            )
            st.write(f"Replace each error latency with block mean (computed in Step 5) + "
                     f"{calculation_params['rt_punishment']} X block SD of correct responses")
        else:
            st.write("Use the reaction time to the correct response when a correction is required. "
                     "Please make sure that your experiment actually recorded the reaction time to the correct "
                     "response when an error is noted for a trial.")

    if st.button("Calculate"):
        if iat_data is None:
            st.error("Please upload your data first.")
            return
        algorithm_method = "improved" if selected_algorithm_index else "conventional"
        algorithm = iat_scorer.IATAlgorithm(algorithm_method, **calculation_params)
        scoring_summary, scored_iat_data = algorithm.process_data(iat_data)
        st.write("Scoring Parameters")
        st.code(algorithm)
        st.write("___")
        st.write("Scored Data Overall Summary")
        st.write(scoring_summary)
        create_downloadable_link(
            scoring_summary.to_csv(index=False),
            "IAT_Data_Scoring_Summary.csv",
            "Download IAT Scoring Summary",
            st
        )
        st.write("___")
        st.write("Scored Data by Response")
        st.write(scored_iat_data)
        create_downloadable_link(
            scored_iat_data.to_csv(index=False),
            "IAT_Data_Score.csv",
            "Download IAT Scores",
            st
        )


def _load_sidebar():
    sidebar.header("Tool List")
    app_tools = ["IAT Generator", "Qualtrics Tools", "IAT Data Scorer"]
    selected_tool = sidebar.radio("Choose Tool", app_tools, 0)
    if selected_tool == app_tools[0]:
        _load_generator()
    elif selected_tool == app_tools[1]:
        _load_qualtrics_tools()
    else:
        _load_scorer()


if __name__ == "__main__":
    _set_width(1200)
    _load_sidebar()



import streamlit as st
from streamlit.report_thread import get_report_ctx

# Use the following session to track data
# Reference: https://gist.github.com/tvst/036da038ab3e999a64497f42de966a92
class SessionState(object):
    def __init__(self, **kwargs):
        """A new SessionState object.
        Parameters
        ----------
        **kwargs : any
            Default values for the session state.
        Example
        -------
        >>> session_state = SessionState(user_name='', favorite_color='black')
        >>> session_state.user_name = 'Mary'
        ''
        >>> session_state.favorite_color
        'black'
        """
        for key, val in kwargs.items():
            setattr(self, key, val)


    @staticmethod
    def get(**kwargs):
        """Gets a SessionState object for the current session.
        Creates a new object if necessary.
        Parameters
        ----------
        **kwargs : any
            Default values you want to add to the session state, if we're creating a
            new one.
        Example
        -------
        >>> session_state = get(user_name='', favorite_color='black')
        >>> session_state.user_name
        ''
        >>> session_state.user_name = 'Mary'
        >>> session_state.favorite_color
        'black'
        Since you set user_name above, next time your script runs this will be the
        result:
        >>> session_state = get(user_name='', favorite_color='black')
        >>> session_state.user_name
        'Mary'
        """
        ctx = get_report_ctx()
        id = ctx.session_id
        return get_session(id, **kwargs)


@st.cache(allow_output_mutation=True)
def get_session(id, **kwargs):
    return SessionState(**kwargs)





from pathlib import Path
import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder

# api_token = "iNKzBVNVAoTMhwnT2amhZRAP4dTBjkEVw9AbpRWg"
# brand_center = "mdanderson.co1"
# data_center = "iad1"
# headers = {"x-api-token": api_token}


class QualtricsTool:
    """Data model to manage Qualtrics-related tools
    Parameters:
    -----------
    api_token: str, the API token for the user
    data_center: str, the data center for the user
    brand_center: str, the brand center for the user
        """
    def __init__(self, api_token=None, data_center=None, brand_center=None):
        self.api_token = api_token
        self.data_center = data_center
        self.brand_center = brand_center

    @property
    def api_headers(self):
        """The default API headers"""
        return {"x-api-token": self.api_token}
    
    @property
    def base_url(self):
        """The default base URL"""
        return f"https://{self.data_center}.qualtrics.com"
    
    @property
    def api_base_url(self):
        """The default base API URL"""
        return f"{self.base_url}/API/v3"
    
    def upload_images_api(self,
                          local_image_folder,
                          library_id,
                          creating_full_url=True,
                          qualtrics_folder=None,
                          filename_pattern="*"):
        """Upload images from the local folder to the Qualtrics server
        :param local_image_folder: str, Path, the local folder containing the images
        :param library_id: str, Qualtrics library ID number
        :param creating_full_url: bool, whether returns the IDs only or the full URLs
        :param qualtrics_folder: str, the Qualtrics Graphics folder for the uploaded images
        :param filename_pattern: str, the pattern using which to select the images for uploading
        :return list[str], the list of image IDs or URLs
        """
        upload_url = f"{self.api_base_url}/libraries/{library_id}/graphics"
        image_urls = list()
        for file in Path(local_image_folder).glob(filename_pattern):
            file_type = Path(file)[1:]
            if file_type not in ("png", "gif", "jpg", "jpeg"):
                raise ValueError("Qualtrics only accepts PNG, GIF, and JPEG images.")
            encoded_fields = {'file': (file.name, open(file, 'rb'), f'image/{file_type}')}
            image_url_id = self._upload_image(encoded_fields, qualtrics_folder, upload_url, file, creating_full_url)
            image_urls.append(image_url_id)
        return image_urls
    
    def upload_images_web(self,
                          image_files,
                          library_id,
                          creating_full_url,
                          qualtrics_folder,
                          image_type):
        """Upload images from the web app to the Qualtrics server
        :param image_files: Bytes, the uploaded bytes data from the web app
        :param library_id: str, Qualtrics library ID number
        :param creating_full_url: bool, whether returns the IDs only or the full URLs
        :param qualtrics_folder: str, the Qualtrics Graphics folder for the uploaded images
        :param image_type: str, the image file type
        :return list[str], the list of image IDs or URLs
        """
        image_urls = list()
        upload_url = f"{self.api_base_url}/libraries/{library_id}/graphics"
        file_count_digit = len(str(len(image_files)))
        for file_i, file in enumerate(image_files, start=1):
            encoded_fields = {'file': (f"image{file_i:0>{file_count_digit}}.{image_type}", file, f'image/{image_type}')}
            image_url_id = self._upload_image(encoded_fields, qualtrics_folder, upload_url, file, creating_full_url)
            image_urls.append(image_url_id)
        return image_urls
    
    def _upload_image(self, encoded_fields, qualtrics_folder, upload_url, file, creating_full_url):
        if qualtrics_folder:
            encoded_fields['folder'] = qualtrics_folder
        mp_encoder = MultipartEncoder(fields=encoded_fields)
        post_request = requests.post(
            upload_url,
            data=mp_encoder,
            headers={'Content-Type': mp_encoder.content_type, **self.api_headers}
        )
        try:
            image_url_id = post_request.json()['result']['id']
        except KeyError:
            raise Exception(f"Failed to upload image {file.name}")
        if creating_full_url:
            image_url_id = f"{self.base_url}/ControlPanel/Graphic.php?IM={image_url_id}"
        return image_url_id
    
    def delete_images(self, library_id, image_url_ids):
        """Delete images from the specified library
        :param library_id: str, the library ID number
        :param image_url_ids: list[str], the image IDs or full URLs
        :return dict, the deletion report"""
        report = dict()
        for image_url_id in image_url_ids:
            if image_url_id.find("=") > 0:
                image_url_id = image_url_id[image_url_id.index("=") + 1:]
            url = f'{self.api_base_url}/libraries/{library_id}/graphics/{image_url_id}'
            delete_response = requests.delete(url, headers=self.api_headers)
            try:
                http_status = delete_response.json()['meta']['httpStatus']
            except KeyError:
                raise Exception(f"Failed to delete image: {image_url_id}")
            else:
                report[image_url_id] = "Deleted" if http_status.startswith('200') else "Error"
        return report
    
    def create_survey(self, template_json):
        """Create the survey using the JSON template
        :param template_json: str in the JSON format, the JSON file for the qsf file
        :return str, the created Survey ID number
        """
        upload_url = f"{self.api_base_url}/survey-definitions"
        creation_response = requests.post(
            upload_url,
            json=template_json,
            headers={**self.api_headers, "content-type": "application/json"}
        )
        try:
            survey_id = creation_response.json()['result']['SurveyID']
        except KeyError:
            raise Exception("Couldn't create the survey. Please check the params.")
        
        return survey_id
    
    def delete_survey(self, survey_id):
        """Delete the survey
        :param survey_id: str, the survey ID number
        :return dict, the deletion report
        """
        report = dict()
        delete_url = f"{self.api_base_url}/survey-definitions/{survey_id}"
        delete_response = requests.delete(delete_url, headers=self.api_headers)
        try:
            http_status = delete_response.json()['meta']['httpStatus']
        except KeyError:
            raise Exception(f"Failed to delete survey: {survey_id}")
        else:
            report[survey_id] = "Deleted" if http_status.startswith('200') else "Error"
        return report
    
    def export_responses(self, survey_id, file_format="csv", data_folder=None):
        """Export responses from the Qualtrics survey"""
        download_url = f"{self.api_base_url}/surveys/{survey_id}/export-responses/"
        download_payload = f'{{"format": "{file_format}"}}'
        download_response = requests.post(
            download_url,
            data=download_payload,
            headers={**self.api_headers, "content-type": "application/json"}
        )
        try:
            progress_id = download_response.json()["result"]["progressId"]
            file_id = self._monitor_progress(download_url, progress_id)
            file_content = self._download_file(download_url, file_id)
        except KeyError:
            raise Exception("Can't download the responses. Please check the params.")
        return file_content
    
    def _monitor_progress(self, download_url, progress_id):
        progress_status = "inProgress"
        while progress_status != "complete" and progress_status != "failed":
            progress_response = requests.get(download_url + progress_id, headers=self.api_headers)
            progress_status = progress_response.json()["result"]["status"]
        return progress_response.json()["result"]["fileId"]
    
    def _download_file(self, download_url, file_id):
        file_url = f"{download_url}/{file_id}/file"
        file_response = requests.get(file_url, headers=self.api_headers, stream=True)
        return file_response.content



import importlib.resources as pkg_resources
import json


class IATTask:
    """Data model to create the IAT task to run on Qualtrics
    
    Parameters
    -----------
    target_positive_concept: str, the positive target
    target_positive_stimuli: list[str], the list of positive target stimuli, words or links to images
    target_negative_concept: str, the negative target
    target_negative_stimuli: list[str], the list of negative target stimuli, words or links to images
    attribute_positive_concept: str, the positive attribute
    attribute_positive_stimuli: list[str], the list of positive attribute stimuli, words or links to images
    attribute_negative_concept: str, the negative attribute
    attribute_negative_stimuli: list[str], the list of negative attribute stimuli, words or links to images
    Refer to the __init__ for other parameters
    """
    templates_names = ["Custom Task", "Flower-Insect", "Gender-Career", "Mexican-US surname"]
    
    def __init__(self,
                 *,
                 target_positive_concept=None,
                 target_positive_stimuli=None,
                 target_negative_concept=None,
                 target_negative_stimuli=None,
                 attribute_positive_concept=None,
                 attribute_positive_stimuli=None,
                 attribute_negative_concept=None,
                 attribute_negative_stimuli=None,
                 **kwargs):
        """Initialization of the IATTask data model
        :param target_stimulus_reference: ["words", "pictures"], how to refer to target stimuli in the instruction
        :param target_stimulus_media: ["text", "image"], what media is used to present the target stimuli
        :param attribute_stimulus_reference: ["words", "pictures"], how to refer to attribute stimuli in the instruction
        :param attribute_stimulus_media: ["text", "image"], what media is used to present the attribute stimuli
        :param left_key_code: str, the key code for entering the left-side response
        :param left_key_name: str, how to refer to the left-side response in the instruction
        :param right_key_code: str, the key code for entering the right-side response
        :param right_key_name: str, how to refer to the right-side response in the instruction
        :param advance_key_code: str, the key code for advancing the task
        :param advance_key_name: str, how to refer to the advance key in the instruction
        :param requires_correction: bool, whether a correction is required when the response is wrong
        :param error_rt_option: ["error", "correct"], when a correction is required, how the reaction time is recorded,
            relative to the initial error trial or the correct trial
        :param auto_advance_delay: int, float, the delay after the incorrect response, it's only valid when a correction
            isn't required
        :param estimated_duration: str, the estimated duration for completing the task
        :param study_name: str, the name for the IAT task, it's required to be set when a survey has multiple IAT
            studies
        :param switch_attributes: bool, whether Block 5 switches attributes or targets
        :param counter_balancing: bool, whether the task should be counterbalanced between the target and attribute
            label pairings
        :param prevent_random: bool, whether you want to prevent random responses (too fast responses)
        :param minimum_allowed_reaction_time: int, float, when prevent_random is True, the minimum allowed reaction time
            below which, the response is considered random
        :param too_fast_response_error_message: str, the error message to be shown when a too fast response is found
        :param overall_instruction_desktop: list[str in the HTML format], the overall instruction for computer users
        :param overall_instruction_mobile: list[str in the HTML format], the overall instruction for mobile users
        :param reminder_instruction_desktop: list[str in the HTML format], the reminder instruction for computer users
        :param reminder_instruction_mobile: list[str in the HTML format], the reminder instruction for mobile users
        :param shows_examples: bool, whether to show the examples
        :param example_duration: int, float, how long each example will be shown, when shows_example is True
        :param example_instruction: list[str in the HTML format], the example instructions
        :param target_label_color: str, the color of the target labels
        :param target_word_color: str, the color of the target stimuli
        :param attribute_label_color: str, the color of the attribute labels
        :param attribute_word_color: str, the color of the attribute stimuli
        :param inter_trial_interval: int, float, the inter trial interval in milliseconds between trials
        :param minimum_preload_image_percent: int, float, the minimum percentage for the image preloading, it only
            applies when the stimuli are images
        :param inter_trial_response_separator: str, the separator between trial responses, which will be saved as
            embedded data fields by Qualtrics
        :param ending_message: str in the HTML format, the message to be shown after the task is over
        :param block_trial_numbers: list, the list of trial numbers for the seven blocks
        :param automatic_responses_delay: int, float, this is for testing purposes, when it's set to be >= the
            minimum_allowed_reaction_time, the automatic correct responses will be entered after the specified delay
        """
        self.target_positive_concept = target_positive_concept
        self.target_positive_stimuli = target_positive_stimuli
        self.target_negative_concept = target_negative_concept
        self.target_negative_stimuli = target_negative_stimuli
        self.attribute_positive_concept = attribute_positive_concept
        self.attribute_positive_stimuli = attribute_positive_stimuli
        self.attribute_negative_concept = attribute_negative_concept
        self.attribute_negative_stimuli = attribute_negative_stimuli
        
        self.target_stimulus_reference = kwargs.get("target_stimulus_reference", "words")
        self.target_stimulus_media = kwargs.get("target_stimulus_media", "text")
        self.attribute_stimulus_reference = kwargs.get("attribute_stimulus_reference", "words")
        self.attribute_stimulus_media = kwargs.get("attribute_stimulus_media", "text")
        
        self.left_key_code = kwargs.get("left_key_code", "KeyF")
        self.left_key_name = kwargs.get("left_key_name", "F")
        self.right_key_code = kwargs.get("right_key_code", "KeyJ")
        self.right_key_name = kwargs.get("right_key_name", "J")
        self.advance_key_code = kwargs.get("advance_key_code", "Space")
        self.advance_key_name = kwargs.get("advance_key_name", "Space")

        self.requires_correction = kwargs.get("requires_correction", True)
        self.error_rt_option = kwargs.get("error_rt_option", "error")
        self.auto_advance_delay = kwargs.get("auto_advance_delay", 300)
        self.estimated_duration = kwargs.get("estimated_duration", "5 minutes")
        self.study_name = kwargs.get("study_name", "")
        self.switch_attributes = kwargs.get("switch_attributes", True)
        self.counter_balancing = kwargs.get("counter_balancing", True)
        self.prevent_random = kwargs.get("prevent_random", True)
        self.minimum_allowed_reaction_time = kwargs.get("minimum_allowed_reaction_time", 250)
        self.too_fast_response_error_message = kwargs.get("too_fast_response_error_message",
                                                          "Too fast! Please choose your answer carefully.")
        
        self.overall_instruction_desktop = self.get_overall_instruction(mobile=False)
        self.overall_instruction_mobile = self.get_overall_instruction(mobile=True)
        self.reminder_instruction_desktop = self.get_reminder_instruction(mobile=False)
        self.reminder_instruction_mobile = self.get_reminder_instruction(mobile=True)
        self.shows_examples = kwargs.get("shows_examples", False)
        self.example_duration = kwargs.get("example_duration", 1500)
        self.example_instruction = self.get_example_instruction()

        self.target_label_color = kwargs.get("target_label_color", "#ffa500")
        self.target_word_color = kwargs.get("target_word_color", "#ffa500")
        self.attribute_label_color = kwargs.get("attribute_label_color", "#003B00")
        self.attribute_word_color = kwargs.get("attribute_word_color", "#003B00")
        
        self.inter_trial_interval = kwargs.get("inter_trial_interval", 250)
        self.minimum_preload_image_percent = kwargs.get("minimum_preload_image_percent", 50)
        self.inter_trial_response_separator = kwargs.get("inter_trial_response_separator", "_")
        self.ending_message = kwargs.get(
            "ending_message",
            "<p style='color: darkgreen'><br /><br />You have completed this task. "
            "<br /><br />"
            "Thank you for your time.<br /><br /></p>"
        )
        if self.ending_message is None:
            self.ending_message = "null"

        self.block_trial_numbers = kwargs.get(
            "block_trial_numbers",
            [20, 20, 20, 40, 20, 20, 40]
        )
        self.automatic_responses_delay = kwargs.get("automatic_responses_delay", 0)

    @staticmethod
    def custom_params(name):
        """Custom parameters for the IAT task with the specified name"""
        pleasant_words = "Joy Happy Laughter Love Friend Pleasure Peace Wonderful".split()
        unpleasant_words = "Evil Agony Awful Nasty Terrible Horrible Failure War".split()
        params = dict()
        if name == IATTask.templates_names[1]:
            params.update(
                target_positive_concept="Flower",
                target_positive_stimuli="Orchid Tulip Rose Daffodil Daisy Lilac Lily".split(),
                target_negative_concept="Insect",
                target_negative_stimuli="Wasp Flea Roach Centipede Moth Bedbug Gnat".split(),
                attribute_positive_concept="Pleasant",
                attribute_positive_stimuli=pleasant_words,
                attribute_negative_concept="Unpleasant",
                attribute_negative_stimuli=unpleasant_words,
                study_name="flower"
            )
        elif name == IATTask.templates_names[2]:
            params.update(
                target_positive_concept="Male",
                target_positive_stimuli="John Paul Mike Kevin Steve Greg Jeff Bill".split(),
                target_negative_concept="Women",
                target_negative_stimuli="Amy Joan Lisa Sarah Diana Kate Ann Donna".split(),
                attribute_positive_concept="Career",
                attribute_positive_stimuli="Executive Management Professional Corporation Salary Office "
                                           "Business Career".split(),
                attribute_negative_concept="Family",
                attribute_negative_stimuli="Home Parents Children Family Cousins Marriage Wedding Relatives".split(),
                study_name="gender"
            )
        elif name == IATTask.templates_names[3]:
            params.update(
                target_positive_concept="American",
                target_positive_stimuli="Smith Johnson Williams Brown Jones Miller Davis".split(),
                target_negative_concept="Mexican",
                target_negative_stimuli="Rodríguez Hernandez García Martínez González López Pérez".split(),
                attribute_positive_concept="Pleasant",
                attribute_positive_stimuli=pleasant_words,
                attribute_negative_concept="Unpleasant",
                attribute_negative_stimuli=unpleasant_words,
                study_name="surname"
            )
        return params

    @staticmethod
    def shared_params():
        """Create the shared parameters for the example IAT tasks"""
        return dict(
            target_stimulus_reference="words",
            target_stimulus_media="text",
            attribute_stimulus_reference="words",
            attribute_stimulus_media="text"
        )

    @staticmethod
    def reset_template(name):
        """Reset the template"""
        shared_kwargs = IATTask.shared_params()
        shared_kwargs.update(IATTask.custom_params(name))
        return IATTask(**shared_kwargs)

    @staticmethod
    def templates():
        """Create the template IAT tasks for the IAT web app"""
        shared_kwargs = IATTask.shared_params()
        templates_kwargs = list()
        for name in IATTask.templates_names:
            template_kwargs = shared_kwargs.copy()
            template_kwargs.update(IATTask.custom_params(name))
            templates_kwargs.append(template_kwargs)
        return [IATTask(**kwargs) for kwargs in templates_kwargs]

    def generate_script(self):
        """Create the script needed for the Qualtrics-IAT experiment"""
        task_stimulus_labels = {
            "p": self.target_positive_concept,
            "n": self.target_negative_concept,
            "+": self.attribute_positive_concept,
            "-": self.attribute_negative_concept
        }
        task_opposite_stimulus_labels = {
            "p": self.target_negative_concept,
            "n": self.target_positive_concept,
            "+": self.attribute_negative_concept,
            "-": self.attribute_positive_concept
        }
        task_attribute_stimulus_type = {
            "name": self.attribute_stimulus_reference,
            "media": self.attribute_stimulus_media
        }
        task_target_stimulus_type = {
            "name": self.target_stimulus_reference,
            "media": self.target_stimulus_media
        }
        task_stimulus_sources = f'{{"p": {self.target_positive_stimuli},' \
                                f'"n": {self.target_negative_stimuli},' \
                                f'"+": {self.attribute_positive_stimuli},' \
                                f'"-": {self.attribute_negative_stimuli}}}'
        task_left_key = {"code": self.left_key_code, "name": self.left_key_name}
        task_right_key = {"code": self.right_key_code, "name": self.right_key_name}
        task_advance_key = {"code": self.advance_key_code, "name": self.advance_key_name}

        task_setup = f"""let task = {{
    overallInstruction: {self.overall_instruction_desktop},
    mobileOverallInstruction: {self.overall_instruction_mobile},
    reminderInstruction: {self.reminder_instruction_desktop},
    mobileReminderInstruction: {self.reminder_instruction_mobile},
    endingMessage: {self.ending_message!r},
    stimulusLabels: {task_stimulus_labels!r},
    oppositeStimulusLabels: {task_opposite_stimulus_labels!r},
    stimulusSources: {task_stimulus_sources},
    attributeStimulusType: {task_attribute_stimulus_type!r},
    targetStimulusType: {task_target_stimulus_type!r},
    blockTrialNumbers: {self.block_trial_numbers},
    leftKey: {task_left_key},
    rightKey: {task_right_key},
    advanceKey: {task_advance_key},
    attributeLabelColor: {self.attribute_label_color!r},
    attributeWordColor: {self.attribute_word_color!r},
    targetLabelColor: {self.target_label_color!r},
    targetWordColor: {self.target_word_color!r},
    showExamples: {str(self.shows_examples).lower()},
    exampleDisplayTime: {self.example_duration},
    exampleInstruction: {self.example_instruction},
    interTrialInterval: {self.inter_trial_interval},
    minimumAllowedReactionTime: {self.minimum_allowed_reaction_time},
    minimumPreloadImagePercent: {self.minimum_preload_image_percent},
    interTrialResponseSeparator: {self.inter_trial_response_separator!r},
    requiresCorrection: {str(self.requires_correction).lower()},
    autoAdvanceDelay: {self.auto_advance_delay},
    tooFastResponseErrorMessage: {self.too_fast_response_error_message!r},
    blockConditions: [],
    studyName: {self.study_name!r},
    switchAttribute: {str(self.switch_attributes).lower()},
    counterBalancing: {str(self.counter_balancing).lower()},
    automaticResponsesDelay: {self.automatic_responses_delay}
}};"""
        standard_script = pkg_resources.read_text("templates", "iat_question_js_code.js")
        return task_setup + '\n' + standard_script

    def generate_template_file(self):
        question_js = self.generate_script()
        json_template = json.loads(pkg_resources.read_text("templates", "iat_survey_template.qsf"))
        for item in filter(lambda x: x == "SurveyElements", json_template):
            for element in filter(lambda x: x["Element"] == "SQ", json_template[item]):
                element["Payload"]["QuestionJS"] = question_js
                break
        dumped_js = json.dumps(json_template)
        return dumped_js
    
    @staticmethod
    def _get_qsf_template():
        return pkg_resources.read_text("templates", "iat_survey_template.qsf")

    @property
    def _stimuli_types(self):
        _stimuli_types = self.attribute_stimulus_reference
        if self.attribute_stimulus_reference != self.target_stimulus_reference:
            _stimuli_types = _stimuli_types + " and " + self.target_stimulus_reference
        return _stimuli_types

    @property
    def _stimuli_types_singular(self):
        _stimuli_types_singular = self.attribute_stimulus_reference[:-1]
        if self.attribute_stimulus_reference != self.target_stimulus_reference:
            _stimuli_types_singular = _stimuli_types_singular + " or " + self.target_stimulus_reference[:-1]
        return _stimuli_types_singular

    def get_overall_instruction(self, mobile):
        paged_instruction = [
            f"<p>In this task, you will be presented with a set of {self._stimuli_types} to classify into groups."
            f"<br /><br />"
            f"The task requires that you classify the {self._stimuli_types} as quickly as you can "
            f"while making as few mistakes as possible."
            f"<br /><br />"
            f"This task takes about {self.estimated_duration}."
            f"<br /><br /></p>"
        ]
        if mobile:
            paged_instruction.extend([
                f"<div>In this task, you'll need to tap the touch screen to answer the questions.<br /><br />"
                f"When the {self._stimuli_types_singular} matches the label on the left side, "
                f"tap the <strong>left</strong> button.<br /><br />"
                f"When the {self._stimuli_types_singular} matches the label on the right side, "
                f"tap the <strong>right</strong> button.</div>",
                "If you're using a mobile device, please hold your device vertically and scroll "
                "the page to ensure that the entire box is visible on the screen."
                "<br /><br />"
                "Note: Don't pull down your page too fast, which may accidentally refresh the page and "
                "you'll lose your progress."
            ])
        else:
            paged_instruction.append(
                f"<div>In this task, you'll need to press {self.left_key_name} and {self.right_key_name} keys on "
                f"your keyboard to answer the questions.</div>"
                f"<div><br />When the {self._stimuli_types_singular} matches the label on the left side, press "
                f"<strong>{self.left_key_name}</strong> key.<br />"
                f"<br />When the {self._stimuli_types_singular} matches the label on the right side, press "
                f"<strong>{self.right_key_name}</strong> key.</div>"
            )
        return paged_instruction

    def get_example_instruction(self):
        return [
            f"<br /><br /><br />"
            f"<p>The following is the list of category labels for this study: "
            f"<strong>{self.target_positive_concept}</strong>, "
            f"<strong>{self.target_negative_concept}</strong>, "
            f"<strong>{self.attribute_positive_concept}</strong>, and "
            f"<strong>{self.attribute_negative_concept}</strong>.<br /><br />" 
            f"You will now see examples for each category. "
            f"The {self._stimuli_types} will be presented, with the categories shown at the top of the screen.",
            f"<br /><br /><br />"
            f"Each {self._stimuli_types_singular} will be shown briefly.<br /><br />"
            f"At this stage, you do not need to make any response.<br /><br />"
            f"Just look at the {self._stimuli_types_singular} carefully.</p>"
        ]

    def get_reminder_instruction(self, mobile):
        if mobile:
            response_element = "button"
            element0 = "Place your device on a desk such that you can tap the screen with your index fingers " \
                       "very quickly."
            element1 = "Position your device properly to enable fast response."
            element2 = f"Two buttons at the bottom will tell you which {self._stimuli_types} go with each button."
            displayed_left_key_name = "left"
            displayed_right_key_name = "right"
        else:
            response_element = "key"
            element0 = f"Place your left index finger on the {self.left_key_name} key " \
                       f"and your right index finger on the {self.right_key_name} key on your keyboard."
            element1 = f"Keep your index fingers on the <strong>{self.left_key_name} and " \
                       f"{self.right_key_name} keys</strong> to enable fast response."
            element2 = f"Two labels at the top will tell you which {self._stimuli_types} go with each key."
            displayed_left_key_name = self.left_key_name
            displayed_right_key_name = self.right_key_name
        if self.requires_correction:
            error_action = f"please correct your answer by pressing the other {response_element}"
        else:
            error_action = "please wait for the task to proceed by itself"
        return [
            f"<ul>"
            f"<li style='text-align: left;'>{element0}</li>"
            f"<li style='text-align: left;'>Press <strong>the {displayed_left_key_name} {response_element} "
            f"with your left "
            f"index finger</strong> if the {self._stimuli_types_singular} fits the category on the left.</li>"
            f"<li style='text-align: left;'>Press <strong>the {displayed_right_key_name} {response_element} "
            f"with your right "
            f"index finger</strong> if the {self._stimuli_types_singular} fits the category on the right.</li>"
            f"<li style='text-align: left;'>If you make a mistake, you will see an error message, "
            f"and {error_action}</li>"
            f"</ul>",
            f"<div style='text-align: center;'><strong>Keep in mind the following points</strong></div>"
            f"<ol>"
            f"<li style='text-align: left;'>{element1}</li>"
            f"<li style='text-align: left;'>{element2}</li>"
            f"<li style='text-align: left;'>Each {self._stimuli_types_singular} has a correct classification. "
            f"Most of these are easy.</li>"
            f"<li style='text-align: left;'>The test gives no results if you go slow. Please try to "
            f"<strong>go as fast as possible!</strong></li>"
            f"<li style='text-align: left;'><strong>Expect to make a few mistakes</strong> because of "
            f"going fast. That's OKAY!</li>"
            f"</ol>"
         ]



from pathlib import Path
import cv2
import numpy as np
import os


def create_word_images(words,
                       target_folder,
                       category_name,
                       font=cv2.FONT_HERSHEY_DUPLEX,
                       font_scale=1,
                       image_size=(400, 400),
                       text_color=(0, 0, 0),
                       fill_color=(255, 255, 255)):
    """Create images for text stimuli as the center
    :param words: the list of words
    :param target_folder: the folder where the images will be saved
    :param category_name: the category name for the words (e.g., Pleasant, Good)
    :param font: the font of the text, by default, Hershey Duplex
    :param font_scale: the size of the font
    :param image_size: the size of the image (width, height), by default, 400 x 400
    :param text_color: the color (rgb values) of the word, by default, black
    :param fill_color: the color (rgb values) of the background, by default, white
    :return: None
    """
    for word in words:
        word_image = _image_with_color(fill_color, image_size)
        _draw_text(word_image, word, font, font_scale, text_color)
        target_path = _create_target_path(target_folder)
        file_destination = target_path / f"{category_name}_{word}.png"
        cv2.imwrite(os.fspath(file_destination), word_image)


def format_images(raw_folder,
                  target_folder,
                  category_name,
                  image_filename_pattern="*",
                  image_size=(400, 400),
                  fill_color=(255, 255, 255)):
    """Format the images to the desired size
    :param raw_folder: the folder where the raw images are saved, JPEG or PNG are preferred, BMP is OK
    :param target_folder: the folder where the created images are saved
    :param category_name: the category name for the images
    :param image_filename_pattern: the image filename pattern, by default, selects all
    :param image_size: the size of the image, default 400 x 400 pixels
    :param fill_color: the color when the image can't cover the desired size, default white
    :return: None
    """
    for image_file in Path(raw_folder).glob(image_filename_pattern):
        target_image = _image_with_color(fill_color, image_size)
        target_width, target_height = image_size
        raw_image = cv2.imread(str(image_file))
        raw_height, raw_width, _ = raw_image.shape
        height_ratio = target_height / raw_height
        width_ratio = target_width / raw_width
        change_ratio = min(width_ratio, height_ratio)
        resized_size = (min(int(raw_width * change_ratio), target_width),
                        min(int(raw_height * change_ratio), target_height))
        resized_image = cv2.resize(raw_image, resized_size)
        offset = (target_width - resized_size[0]) // 2, (target_height - resized_size[1]) // 2
        _paste(resized_image, target_image, offset)
        target_path = _create_target_path(target_folder)
        file_path = target_path / f"{category_name}_{image_file.stem}.png"
        cv2.imwrite(os.fspath(file_path), target_image)


def _create_target_path(target_folder):
    target_path = Path(target_folder)
    if not target_path.exists():
        target_path.mkdir(parents=True, exist_ok=True)
    return target_path


def _paste(small_image, large_image, position=(0, 0)):
    """
    Paste a small image on a larger image
    :param small_image: the smaller image
    :param large_image: the larger image
    :param position: the position where the small image is pasted
    :return: None
    """
    x_offset, y_offset = position
    large_image[y_offset:y_offset+small_image.shape[0], x_offset:x_offset+small_image.shape[1]] = small_image


def _image_with_color(color, size):
    """
    Create an image with a solid color
    :param color: the color (rgb values) of the image
    :param size: the size of the image
    :return: the image that is created
    """
    width, height = size
    color_image = np.ones(shape=[height, width, 3], dtype=np.uint8)
    for color_index, color_value in enumerate(color):
        color_image[:, :, color_index] = color_value
    return color_image


def _draw_text(image, text, font, font_scale, color, x=None, y=None):
    """
    Draw text on an image
    :param image: the image that is to add an text
    :param text: the text that is to add to the image
    :param font: the font of the text
    :param font_scale: the font scale
    :param color: the color of the text
    :param x: the horizontal coordinate (x) for the text
    :param y: the vertical coordinate (y) for the text
    :return: None
    """
    thickness = 1
    image_height, image_width = image.shape[0], image.shape[1]
    text_width, text_height = cv2.getTextSize(text, font, font_scale, thickness)[0]
    if x is None:
        x = (image_width - text_width) // 2
    if y is None:
        y = (image_height + text_height) // 2
    cv2.putText(image, text, (x, y), font, font_scale, color, thickness, cv2.LINE_AA)



import setuptools

with open("README.md") as file:
    read_me_description = file.read()

setuptools.setup(
    name="qualtrics_iat",
    version="0.1",
    author="Yong Cui",
    author_email="ycui1@mdanderson.org",
    description="A toolset for creating and scoring Qualtrics-based IAT experiments",
    install_reqruies=['pandas', 'streamlit', 'openpyxl', 'xlrd'],
    long_description=read_me_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ycui1-mda/qualtrics_iat",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)
